
#include <string>
#include <vector>
#include "stoch.hh"
#include <ctime>
#include <cstdlib>
#include <strstream>
#include <fstream>
#include <time.h>

//function to display help
void Usage ()
{
  cout << "Usage:"  << endl;
  cout << "./stoch <species data file> <reaction data file>" 
    " <end time> "
    " <interval for output> <output file name> <number of replicates>"
    << endl;
  
  cout << "If no replicates is specified, only one will be made." << endl;
  
}

// function to initialize the average output file
void initializeAverageOutput (speciesTable& sT, char*& fileName)
{
  ostrstream os;
  os << fileName << ".average" << ends;
  ofstream aveOut (os.str());
  os.freeze(0); 
  aveOut << "Time\t";
  //put a header in the above output file
  for (int p=0; p < sT.speciesVector.size(); p++)
    {
      if (sT.speciesVector[p]->output)
	{
	  aveOut << sT.speciesVector[p]->speciesName << "(-stdv)\t"
		 << sT.speciesVector[p]->speciesName << "(mean)\t"
		 << sT.speciesVector[p]->speciesName << "(+stdv)\t";
	}
    }
  aveOut << endl;
  
  aveOut << "0\t"; // inital time is 0
  sT.averator (aveOut);
  aveOut << endl;
}

void outputAverageData (char*& fileName, speciesTable& sT, double outputTime)
{
  ostrstream os;
  os << fileName << ".average" << ends;
  ofstream aveOut (os.str(), ios::app);
  os.freeze(0); 
  aveOut << outputTime << "\t";
  sT.averator (aveOut);
  aveOut<< endl;
}

// function to initialize the replicate output files
void initializeReplicateOutput (int replicate, char*& fileName, speciesTable& sT)
{
  for (int i = 0; i < replicate; i++)
    {
      ostrstream os;
      os << fileName << (i+1) << ".out" << ends;
      os.freeze(0);
      
      ofstream out_file (os.str()); 

      //the following bit of code puts a header in the output file 
      out_file << "Time\t";

      for (int p=0; p < sT.speciesVector.size(); p++)
	{
	  if (sT.speciesVector[p]->output)
	    out_file << sT.speciesVector[p]->speciesName << "\t";
	}
      out_file << endl;
      
      //      dump to the output files the initial conditions
        out_file << "0\t" ; //time is 0
      for (int e = 0; e < sT.speciesVector.size(); e++)
  	{
  	  if (sT.speciesVector[e]->output)
  	    out_file << sT.speciesVector[e]->aAmountSpecies[0] << "\t";
  	}
      out_file << endl;
      out_file.close();
    }
}


// function to output replicate data
void outputReplicateData (speciesTable& sT, prioriQtable& vprQ, char*& fileName, int w)
{
  ostrstream os;
  os << fileName << (w+1) << ".out" << ends;
  
  ofstream out_file (os.str(), ios::app); 
  out_file << vprQ.pqVector[w]->totalTime << "\t";
  
  for (int e = 0; e < sT.speciesVector.size(); e++)
    {
      if (sT.speciesVector[e]->output)
	out_file << sT.speciesVector[e]->aAmountSpecies[w] << "\t";
    }
  
  out_file << endl;
  os.freeze(0);
  out_file.close();
}


// function to check the number of arguments passed in
int checkArg (int argc, char* argv[])
{
  if ((argc == 6) || (argc == 7)) 
    {
      cout << "HERE!";
      
      return 0;
    }
  else 
    {
      Usage ();
      return 1;
    }
  
};

//  Welcome to the stochastirator!  
//  This program models the stochastic events for a biological system where
//  small numbers of molecules are used.
//  It is designed to use two data input files:
//      1.  The species data file.  This should contain every species 
//          used paired with its initial amount.
//          Example of 3 enteries:
//
//            Na 100
//            Cl 100
//            NaCl 0
//
//      2.  The reaction data file.  This should contain every reaction to be
//          used in the simulation.  
//          Example areactions:
//
//             Rxn Na+Cl->NaCl
//             Subs Na Cl
//             Prods NaCl
//             Rate 10
//
//             Note:  the line headers -- Rxn, Subs, Prods, Rate -- are
//             necessary for file parsing.  Subs and Prods can be delineated
//             by space or by line:
//             
//             Rxn Na+Cl->NaCl
//             Subs
//             Na
//             Cl
//             Prods NaCl
//             Rate 10
//
//  From these two files, three major data sets are created.
//  1. The Reaction Data Table class
//     This table is a vector of pointers to a class, reaction data. This also
//     contains member functions that are used to affect all reaction data 
//     such as initializing all the time for next reaction execution.
//
//       a. Reaction Data base class and reaction type derived classes
//
//          The reaction data base class stores all the information about one
//          reaction -- name, substrates, products, reaction rate, reaction
//          graph (more on this guy later), time for the next reaction
//          execution, the value of h*c, and a few more flags and values.  
//          Member functions for this class include calculating the time for 
//          the next reaction execution, decreasing products, and increasing 
//          substrates
//
//          reaction type derived classes:
//          Since each reaction type (evaluated from the number and type of 
//          substrates) causes h to be calculated differently, I added derived
//          type classes from the reaction data base class to aid in data
//          structuring.  Each derived class contains all the info from the 
//          reaction data base class, plus an added member function for the 
//          calculation of h.
//
//  2. The species table class.
//     This table is a vector of pointers to a species data class.
//     Member functions are for creating the table from a data file. 
//
//     a. species class
//        stores two pieces of information:  the name of the species, and 
//        the amount of the species.
//
//  3.  The prioity queue class.
//      This is simply an array (called timeArray) of pointers to reaction data.
//      It contains member functions used to sort the array as a priority 
//      queue based on the value of the next time for reaction execution 
//      stored in the reaction data.
//
//
//  The overall flow of the program:
//    1. the species table is initialized
//       a. open species data file
//       b. read from file the species name and amount.
//       c. create a new species class for b.
//       d. add c. to species table vector
//       e. repeat b.->d. until eof
//
//    2. the reaction data table is initialized
//       a. open reaction data file
//       b. read from file the rxn name, substrates, products, and rate
//       c. use the substrates to figure out the reaction type
//       d. create a new derived reaction type class for that reaction type
//            a'.  pass in rxn name, rate, subs, and prods
//            b'.  name and rate are stored values
//            c'.  a vector of pointers to species class in species table is
//                 created for subs and prods
//       e. repeat b->d until eof. 
//       f. for each reaction data, create a reaction graph
//          Note, the reaction graph is a vector of pointers to other
//          reaction data classes. These reaction data are dependent upon
//          the substrates and/or products of the self reaction data.
//          This is used to optimize the number of reactions that must be 
//          updated if the self reaction is executed (instead of updating 
//          all the reactions.)
//       g. The time for next reaction execution is calculated for all
//          reaction data.  
//    3. The prioity queue is initialized.
//      a. For each member of the timeArray, a pointer to a member of 
//         the reaction data table is created.
//      b. The array is sorted based off the time for next reaction execution
//         stored in the reaction data member.
//    4. Execute the reaction with the smalled time for next reaction execution
//       (at position 0 in the priority queue.)
//      a. decrease the amount of substrate species
//      b. increase the amount of product species
//      c. calculate new next rxn time for self
//      d. sort priority queue
//      e. calculate and sort new/adjusted rxn time for those in the reaction graph
//    5. Dump species amount to data file.
//    6. Repeat 4->5 till finished.

int main (int argc, char* argv[])
{
  double startTime=time(0);
  cout << endl << "Starting the STOCHASTIRATOR V0.55. . ." << endl << endl;
  cout << "argc: " << argc << endl;
  
  char* speciesFileName = "";
  char* rxnFileName = "";
  char* cOutputInterval = "";
  float outputInterval = 0;
  char* fileName = "";
  char* cReplicate = "";
  int replicate = 0;
  double endTime = 0;
  
  //check command line for proper use.  If not, exit. 
  if (checkArg(argc, argv) == 1)
    return 1;

  //get a a couple of parameters from the command line . . .
  speciesFileName = argv[1];
  rxnFileName = argv[2];

  //get time interval for dumping the data  
  
  endTime = atof(argv[3]);
  
  cout << "End time:  " << endTime << endl << endl;
  
  outputInterval = atof(argv[4]);
  
  cout << "Output Interval:  " << outputInterval << endl << endl;
  
  // name for the output file
  
  fileName = argv[5];
   
  cout << "Output file name base:  " << fileName << endl << endl;
  
  // number of replications
  
  if (argv[6])
    replicate = atoi(argv[6]);
  else
    replicate = 1; //default

  cout << "Number of replicates:  " << replicate << endl << endl;
  
 // open a file for recording switching events
  ostrstream swit;
  swit << fileName << ".switch" << ends;
  ofstream test (swit.str());
  swit.freeze(0);  
  //open for output a file to dump diagnostic data in
  ostrstream diag;
  diag << fileName << ".diag" << ends;
  ofstream diagnos (diag.str());
  diag.freeze(0);  
  
  diagnos << "Species data file:  " << speciesFileName << endl;
  diagnos << "Reaction data file:  " << rxnFileName << endl;
  diagnos << "Output Interval:  " << outputInterval << endl;  
  diagnos << "Output file name base:  " << fileName << endl;
  diagnos << "Number of replicates:  " << replicate << endl;
  diagnos << "End time:  " << endTime << endl;

//generate a seed value for the random number generator and save it.
  unsigned int seed = (unsigned int)time(0);
  srand48(seed); //seed random number
  diagnos << "Random number seeded with:  " << seed <<endl;  

  // open the reaction data file
  ifstream rxnIn(rxnFileName);
  if (!rxnIn.is_open())
    {
      cout << "Out Species file, " << rxnFileName << ", is not open." << endl;
      cout << "Check file name and permissions." << endl;
      return 1;
    }
  
  //open the species data file
  ifstream speciesIn(speciesFileName);
  if (!speciesIn.is_open())
    {
      cout << "Out Species file, " << speciesFileName << ", is not open." << endl;
      cout << "Check file name and permissions." << endl;
      return 1;
    }
  
  // define and create the species table
  
  speciesTable sT (replicate);
  sT.readSpeciesFile(speciesIn);
  sT.dumpSpeciesTable(diagnos); // check initialization
  
  // define and create the reaction data table
  reactionDataTable rDT (replicate);
  rDT.readReactionDataFile (rxnIn, sT);
  // make the reaction graph for all the reations
  rDT.makeReactionGraph();
  // dump the reaction data table to the diagnostic outfile
  rDT.dumpReactionDataTable(diagnos);
  //generate the first round of times to execute
  rDT.initializeTimes();
  //define and create the priority queue table and, sort each prioity queue 
  prioriQtable vprQ(replicate, rDT);
  vprQ.dumpPrioriQtable(diagnos); //check prioriQ initialization
  //initialize the outputTime.
  double outputTime = 0;
  
  //need to open an output file for averages, etc.
  if (replicate != 1)
    initializeAverageOutput (sT, fileName);
  
  // if (outputReplicates)
    initializeReplicateOutput (replicate, fileName, sT);
  
  //the reaction execution loop.  Two events will signal the termination of this loop
  // 1.  the total number of interations has been completed
  // 2.  a reaction with no substrate species left has been executed signalling that 
  //     either something very wrong happened in the priority queue (if a reaction has
  //     no substrate then its execution time should be infinity) or that simply all
  //     the substrate for all the reactions has been used up (all reactions have 
  //     infinite times for next execution
  bool crash = 0;
  int i = 0;
  double iteration = 0;
  while  (outputTime <= endTime)
    //while  (vprQ.pqVector[0]->totalTime <= endTime)
    {
      //lame progress meter
      if (i % 100 == 0)
  	{
  	  cout << outputTime/endTime*100;
  	  cout << " percent completed." << endl;
  	}
      
      outputTime = outputTime + outputInterval;
      
      
      for (int w = 0; w < replicate; w++)
	{
	  //  Interation keeps track of the number of time a replicate had to execute
	  //  before advancing its totalTime to be greater than the outputTime.
	  //iteration = 0;
	  
	  while ((vprQ.pqVector[w]->totalTime <= outputTime) && (vprQ.pqVector[w]->totalTime <= endTime)) 
	  {	      
	      iteration++;
	      // Execute the reaction in the root (position 0) of the priori Q.
	      //test << "Total Time:  " << vprQ.pqVector[w]->totalTime << endl;
	      
	      //vprQ.dumpPrioriQtable (test);
	      //sT.dumpSpeciesTable (test);
	      //test << "end reaction" << endl << endl << endl;
	      
	      vprQ.pqVector[w]->executeReaction (0, w, test);
	      //    if ((vprQ.pqVector[w]->totalTime > endTime) || (vprQ.pqVector[w]->totalTime < 0))
	      //	{
	      //	  vprQ.dumpPrioriQtable(diagnos);
	      //	  crash = 1;
	      //	} 
	      }
	  // If no reaction events have occurred, skip to the next replicate.
	  // Otherwise, time to dump some data!
	  if ((iteration != 0))
	    {
	      outputReplicateData (sT, vprQ, fileName, w);
	    }
	}
            
      if (replicate != 1)
	outputAverageData (fileName, sT, outputTime);
     
      if (crash)
	{
	  break;
	}
      i++;
    }

  vprQ.dumpPrioriQtable(diagnos);
  
  cout << "time limit " << endTime << " reached, crashing out." << endl;
  diagnos << "time limit " << endTime << " reached, crashing out." << endl;
  diagnos << iteration << " reaction events completed." << endl;
  
  double finishTime = (time(0)-startTime);
  
  if (finishTime < 60)
    diagnos << "Run time:  " << finishTime << " seconds."<< endl;
  else if (finishTime < 3600)
    diagnos << "Run time:  " << finishTime/60 << " minutes." << endl;
  else 
    diagnos << "Run time:  " << finishTime/3600 << " hours." << endl;
    
  return 0;
}










